/*
 * Audiosync
 *
 * This is a quick and dirty utility to resync DV audio in a quicktime file.
 *
 * Most consumer camcorders run with what's known as unlocked audio.  This
 * means the audio sampling clock is not locked to the video frame clock. 
 * There are two variations to unlocked audio.  The first (which doesn't
 * really need correcting) is where the clocks are allowed to drift over the
 * short-term; however, over the long-term Na/Ar == Nv/Vr.  Here Na,Nv are
 * the number of audio/video samples and Ar,Vr are the nominal audio/video
 * sampling/frame rates.  With this type of unlocked audio, the length of
 * video and audio tracks in a captured quicktime file will generally
 * be in agreeance when converted to elapsed time.
 *
 * The second type of unlocked audio is where the audio and video clocks
 * are completely decoupled.  This means that over time the two drift with
 * respect to each other.  The end result is that a captured quicktime
 * file will have different time lengths for video and audio tracks when
 * the respective sampling rates are assumed to be their nominal values.
 *
 * In addition to Quicktime files, this program can also perform resampling
 * on any audio file supported by libsndfile in a similar way as for
 * Quicktime files.  Again, this process is designed to resynchronise a
 * digital audio file with some other digital clock source.
 *
 * This utility was written by Jonathan Woithe and is (c) 2005 Jonathan
 * Woithe.  The initial skeleton was qtinfo.c, provided by libquicktime
 * 0.9.3.  This was written by by Elliot Lee <sopwith@redhat.com>.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
 * Public License for more details.
 *       
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* Support large files (essential for video work) */
#define _GNU_SOURCE
#define _FILE_OFFSET_BITS 64

/* Ensure QT_INC_IN_QUICKTIME is defined.  Default to 0 since this is the
 * appropriate setting for recent libquicktime releases.  It should be
 * set to 1 only if the libquicktime include files are found in 
 * .../include/quicktime/.
 */
#ifndef QT_INC_IN_QUICKTIME
#define QT_INC_IN_QUICKTIME 0
#endif

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <samplerate.h>
#include <sndfile.h>

#if QT_INC_IN_QUICKTIME == 0 
  #include <lqt/quicktime.h>
  #include <lqt/lqt.h>
#else
  #include <quicktime/quicktime.h>
  #include <quicktime/lqt.h>
#endif

#define VERSIONSTR "0.4"

/* ======================================================================== */
/* Input and output audio buffers */

#define BUFFER_LEN 4096

static float input[BUFFER_LEN], tmp_audio[BUFFER_LEN];
static float output[BUFFER_LEN];

static float **input_channel_ptr = NULL;

/* ======================================================================== */

/* A structure used to abstract details of different source file types */
typedef struct Input_file_info {
  union {
    SNDFILE     *sndfile;
    quicktime_t *qtfile;
  } input_file;
  signed int nominal_input_samplerate;
  signed int output_samplerate;
  float init_gain;
  signed int channels;
  signed int input_n_samples;
  signed int final_n_samples;
  double resample_ratio;
  unsigned int filetype;
} Input_file_info;

/* File types used in Input_file_info::filetype */
#define INPUT_NONE      0
#define INPUT_QT        1
#define INPUT_SNDFILE   2

/* ======================================================================== */

/* Other global variables */
signed int cwd_output = 0;

signed int unnorm_flt = 0;

/* ======================================================================== */

static signed int file_exists(char *filename) {
/*
 * Returns 1 if the given file exists and is readable.
 */
FILE *f = fopen(filename,"r");
signed int res = (f!=NULL);
  if (res)
    fclose(f);
  return res;
}
/* ======================================================================== */

static signed int qtfile_probe(char *filename, Input_file_info *info) {
/*
 * Probes the supplied file to see if it is in a format suitable for this
 * program and whether any work needs to be done on it.  Calculates the
 * resampling ratio which will be used with libsamplerate to do the actual
 * rate conversion and returns this ratio along with a libquicktime handle
 * to the source.  In the event of a problem -1 is returned.  If the file is
 * otherwise ok but there's no resync required, 0 is returned.
 */
quicktime_t* qtfile;
signed int i;
signed int n_audiotracks, n_videotracks;
double audio_time=0.0, video_time=0.0;
signed int audio_ok=1, video_ok=1;
double resample_ratio = 0.0;
  if (!quicktime_check_sig(filename))
    return -1;
  qtfile = quicktime_open(filename, 1, 0);
  if (!qtfile) {
    return -1;
  }
  info->filetype = INPUT_QT;

  printf("\nFile %s is a Quicktime file:\n", filename);
  n_audiotracks = quicktime_audio_tracks(qtfile);
  audio_ok = n_audiotracks==1;
  if (lqt_is_avi(qtfile)) {
    printf("  File is AVI: this has not been tested\n");
  }
  printf("  %d audio tracks.  %s.\n", n_audiotracks,
    audio_ok?"Good":"Only files with 1 audio track supported");

  for (i = 0; i < n_audiotracks; i++) {
    printf("    %d channels, %d bits, nominal sample rate %ld, length %ld samples\n",
      quicktime_track_channels(qtfile, i),
      quicktime_audio_bits(qtfile, i),
      quicktime_sample_rate(qtfile, i),
      quicktime_audio_length(qtfile, i));
    if (!quicktime_supported_audio(qtfile, i)) {
      printf("      Track codec not supported: can't do anything.\n");
      audio_ok=0;
    } else {
      printf("      Track is uses supported codec.  Good.\n"); 
    }
    printf("      Length in seconds: %g\n",audio_time=
      (double)quicktime_audio_length(qtfile, i)/quicktime_sample_rate(qtfile, i));
  }

  n_videotracks = quicktime_video_tracks(qtfile);
  video_ok = n_videotracks==1;
  printf("  %d video tracks.  %s.\n", n_videotracks,
    video_ok?"Good":"Only files with 1 video track supported");
  for (i = 0; i < n_videotracks; i++) {
    printf("    %dx%d, depth %d, rate %f, length %ld frames\n",
      quicktime_video_width(qtfile, i),
      quicktime_video_height(qtfile, i),
      quicktime_video_depth(qtfile, i),
      quicktime_frame_rate(qtfile, i),
      quicktime_video_length(qtfile, i));
    if (!strcmp(quicktime_video_compressor(qtfile, i),QUICKTIME_DV)) {
      printf("      Video is DV.  Good.\n");
    } else {
      printf("      Video is not DV.  Nothing to do.\n");
      video_ok=0;
    }
    printf("      Length in seconds: %g\n",video_time=
      (double)quicktime_video_length(qtfile, i)/quicktime_frame_rate(qtfile, i));
  }

  if (audio_ok && video_ok) {
    /* Allow command line ratio to override ratio implied by file contents */
    if (info->resample_ratio>0.0 || 
        (info->output_samplerate>0 && info->output_samplerate!=quicktime_sample_rate(qtfile,0))) {
      signed int output_f = info->output_samplerate==0?quicktime_sample_rate(qtfile,0):info->output_samplerate;
      if (info->resample_ratio>0.0)
        resample_ratio = info->resample_ratio;
      printf("  Requested resample ratio: %0.6f\n",resample_ratio);
      printf("  Actual input audio frequency implied by resampling ratio: %0.4f Hz\n",
        (float)output_f/info->resample_ratio);
    } else
    if (audio_time==video_time)
      printf("  Audio and video are the same length.  Nothing to fix.\n");
    else {
      /* This is the ratio for libsamplerate.  It's new/old. What we want
       * is exactly the nominal sample rate given in the file header. 
       */
      resample_ratio = quicktime_sample_rate(qtfile,0)*video_time /
        (double)quicktime_audio_length(qtfile, 0);
      printf("  Resample ratio: %0.6f\n",resample_ratio);
      printf("  Actual input audio sample rate relative to video frame rate: %0.4f Hz\n",
        (float)quicktime_audio_length(qtfile, 0)/video_time);
    }
  } else {
    resample_ratio = -1.0;
  }

  if (resample_ratio <= 0) {
    quicktime_close(qtfile);
  } else {
    info->input_file.qtfile = qtfile;
    /* Take audio details from the first audio track */
    info->nominal_input_samplerate = quicktime_sample_rate(qtfile,0);
    info->channels = quicktime_track_channels(qtfile, 0);
    info->input_n_samples = quicktime_audio_length(qtfile, 0);
    info->final_n_samples = (long long signed int)quicktime_sample_rate(qtfile,0)*
      quicktime_video_length(qtfile,0)/quicktime_frame_rate(qtfile,0);
    info->resample_ratio = resample_ratio;

    /* Set up channel buffer pointer array */
    input_channel_ptr = calloc(info->channels,sizeof(float *));
    for (i=0; i<info->channels; i++) {
      input_channel_ptr[i] = tmp_audio + (i*BUFFER_LEN/info->channels);
    }
  }
  return resample_ratio > 0.0;
}
/* ======================================================================== */

static signed int sndfile_probe(char *filename, Input_file_info *info) {
/*
 * Probes the given file to see if it's readable using sndfile library.
 * Returns -1 on error, 0 if the file is OK but no sync is required, and
 * 1 if everything is in place to support a resync.
 */
SNDFILE *file;
SF_INFO fileinfo;
  file = sf_open(filename,SFM_READ,&fileinfo);
  if (file==NULL) {
    return -1;
  }
  info->filetype = INPUT_SNDFILE;
  info->input_file.sndfile = file;
  info->nominal_input_samplerate = fileinfo.samplerate;
  info->channels = fileinfo.channels;
  info->input_n_samples = fileinfo.frames;
  info->final_n_samples = 0;
  printf("\nFile %s is an audio file:\n"
         "  1 audio tracks.  Good.\n"
         "    %d channels, nominal sample rate %d, length %d samples\n"
         "      Length in seconds: %0.4f\n",
    filename,
    info->channels,info->nominal_input_samplerate,info->input_n_samples,
    (float)(info->input_n_samples)/info->nominal_input_samplerate);
  if (info->resample_ratio<=0.0) {
    printf("  No resample ratio requested: nothing to do\n");
    return 0;
  } else
  if (info->resample_ratio==1.0) {
    printf("  Requested resample ratio is 1.0: nothing to do\n");
    return 0;
  } else {
    signed int output_f = info->output_samplerate==0?info->nominal_input_samplerate:info->output_samplerate;
    info->final_n_samples = info->input_n_samples*info->resample_ratio;
    printf("  Requested resample ratio: %0.6f\n",info->resample_ratio);
    printf("  Actual input audio frequency implied by resampling ratio: %0.4f Hz\n",
      (float)output_f/info->resample_ratio);
    printf("  New track length: %d samples (%g seconds)\n",
      info->final_n_samples,(float)info->final_n_samples/output_f);
  }
  return 1;
}
/* ======================================================================== */

static char *output_filename(char *srcname) {
/*
 * Mangles the source filename to make an output filename which is hopefully
 * suitable.
 */

/* Name is oversized by 7 (for "-synced") +4 (for ".wav" in the event the
 * original has no extension).
 */
char *name = malloc(strlen(srcname)+1+7+4);
char *ext;
  if (cwd_output && (ext=strrchr(srcname,'/'))!=NULL) {
    /* Output to current directory; therefore just copy the filename
     * portion of a path if given.
     */
    strcpy(name,ext+1);
  } else {
    strcpy(name,srcname);
  }

  /* Point to the start of the file extension.  If there is no extension,
   * point to the end of the filename.  Don't get confused by filenames
   * without extensions which have dots elsewhere in their path.
   */
  ext = strrchr(name,'.');
  if (ext==NULL || ext<strrchr(name,'/')) {
    ext = name+strlen(name);
  }
  *ext = 0;
  strcat(ext,"-synced.wav");
  return name;
}
/* ======================================================================== */

static float apply_gain (float * data, long frames, int channels, float max, float gain) {
/*
 * Applies the given gain factor to all data in the given array.  If any 
 * post-gain data exceeds max, the new maximum data is returned; otherwise
 * max is returned.
 */
long k ;
  for (k=0; k < frames*channels; k++) {
    data[k] *= gain;
    if (fabs(data[k]) > max)
      max = fabs(data[k]);
  }
  return max ;
}
/* ======================================================================== */

static void qt_read_data(Input_file_info *info, SRC_DATA *src_data) {
/*
 * Reads audio data from a quicktime file into the input[] array.  The
 * libquicktime API doesn't return interleaved samples, so this function
 * interleaves them before returning; this is required for later use
 * of the data by libsamplerate.
 */ 
signed int i, j;
signed long long int pos, curr_pos;
  pos = quicktime_audio_position(info->input_file.qtfile,0);

  if (lqt_decode_audio_track(info->input_file.qtfile,NULL,input_channel_ptr,BUFFER_LEN/info->channels,0))
    src_data->input_frames = 0;
  else {
    curr_pos = lqt_last_audio_position(info->input_file.qtfile,0);
    if (curr_pos > info->input_n_samples)
      src_data->input_frames = info->input_n_samples-pos;
    else
      src_data->input_frames = curr_pos-pos;
  }

  /* Data read into tmp_audio via the input_channel_ptr[] pointer array is
   * not interleaved; libsamplerate requires interleaved data, so interleave
   * it before returning, putting the result into the input[] array.
   */
  for (i=0; i<info->channels; i++) {
    for (j=0; j<src_data->input_frames; j++) {
      input[info->channels*j+i] = tmp_audio[j];
    }
  }
}
/* ======================================================================== */

static signed int do_resample(Input_file_info *info, char *output_name) {
/*
 * Resamples the audio from the given quicktime file according to the
 * ratio supplied.
 */
SNDFILE *outfile = NULL;
SF_INFO sfinfo;
SRC_STATE *src_state = NULL;
SRC_DATA src_data;
signed int channels;
signed long long int audio_length;
signed long long int final_audio_length;
int error=0, complete=0, maxed;
float max = 0.0, gain = 1.0;
sf_count_t output_count=0;

  memset(&sfinfo,0,sizeof(sfinfo));

  /* Override initial gain setting if requested */
  if (info->init_gain > 0.0) {
    gain = info->init_gain;
  }

  /* Override output sample rate if requested */
  if (info->output_samplerate!=0)
    sfinfo.samplerate = info->output_samplerate;
  else
    sfinfo.samplerate = info->nominal_input_samplerate;

  printf("  Resampling: output to %s at %d Hz, gain=%g\n",output_name,sfinfo.samplerate,gain);

  audio_length = info->input_n_samples;
  final_audio_length = info->final_n_samples;

  /* Take channel count from original source */
  channels = info->channels;

  sfinfo.channels = info->channels;
  if (!unnorm_flt) {
    sfinfo.format = SF_FORMAT_WAV | SF_FORMAT_PCM_16;
  } else {
    sfinfo.format = SF_FORMAT_WAV | SF_FORMAT_FLOAT;
  }

  while (!error && !complete) {
    /* Initialise the samplerate converter.  Strange things happen if
     * this is allocated outside this loop and reused for subsequent
     * resampling attempts. */
    if ((src_state = src_new(SRC_SINC_BEST_QUALITY,channels,&error)) == NULL) {
      printf ("  Could not create sample rate converter: %s\n",src_strerror(error));
      return -1;
    }
    src_data.src_ratio = info->resample_ratio;
    src_data.data_out = output;
    src_data.output_frames = BUFFER_LEN / channels;

    maxed = 0;
    max = -1.0;
    src_data.data_in = input;
    src_data.end_of_input = 0; /* Set this later */
    /* Start with zero to force load in while loop */
    src_data.input_frames = 0;
    /* Ensure we're at the beginning of the file */
    switch (info->filetype) {
      case INPUT_QT:
        quicktime_set_audio_position(info->input_file.qtfile,0,0);
        break;
      case INPUT_SNDFILE:
        sf_seek(info->input_file.sndfile,0,SEEK_SET);
        break;
    }
    /* Using the above information, open the output file */
    if ((outfile = sf_open(output_name,SFM_WRITE,&sfinfo)) == NULL) {
      printf("  Could not open %s for writing\n",output_name);
      src_state = src_delete(src_state);
      return -1;
    }
    fflush(stdout);
    while (!error && !complete && !maxed) {
      /* If the input buffer is empty, refill it */
      if (src_data.input_frames==0 && !src_data.end_of_input) {
        switch (info->filetype) {
          case INPUT_QT: qt_read_data(info,&src_data); break;
          case INPUT_SNDFILE: 
            src_data.input_frames = sf_readf_float(info->input_file.sndfile,
              input,BUFFER_LEN/channels);
            break;
        }
        src_data.data_in = input;
        /* The last read will not be a full buffer, so end_of_input */
        if (src_data.input_frames < BUFFER_LEN/channels) {
          src_data.end_of_input = SF_TRUE;
        }
      }
      if ((error = src_process(src_state,&src_data))) {
        printf("  Error during resampling: %s\n",src_strerror(error));
      }
      if (!error && src_data.end_of_input && src_data.output_frames_gen==0) {
        complete = 1;
      }

      max=apply_gain(src_data.data_out,src_data.output_frames_gen,channels,
            max,gain);
      if (max>1.0 && !unnorm_flt) {
        gain = gain/max;
        printf("  Resampling produced clipping after sample %lld (%g sec)\n"
               "  Restarting with a gain of %g\n",
          output_count,(float)output_count/sfinfo.samplerate,gain);
        error = complete = 0;
        output_count = 0; 
        maxed=1;
        sf_close(outfile);
        src_state = src_delete(src_state);
      }
      /* Write output. */
      if (!maxed) {
        if (!complete) {
          output_count += src_data.output_frames_gen;
          /* Roundoff can sometimes cause one too many samples to be
           * generated.  If this has occured just drop the last one so the
           * length is exactly matching the video length.
           */
          if (output_count > final_audio_length) {
            src_data.output_frames_gen -= (output_count-final_audio_length);
            output_count = final_audio_length;
          }
          sf_writef_float(outfile,output,src_data.output_frames_gen) ;
        }
        src_data.data_in += src_data.input_frames_used * channels;
        src_data.input_frames -= src_data.input_frames_used;
      }
    }
  }

  if (!error) {
    printf("  Successfully wrote audio track: %lld samples, %g seconds\n",
      output_count, (float)output_count/sfinfo.samplerate);
    if (output_count < final_audio_length) {
      printf("  Warning: %lld samples needed to satisfy request, %lld generated\n",
        final_audio_length, output_count);
    }
    if (max > 1.0) {
      printf("  Note: Maximum output amplitude was %g\n",max);
    }
  }
  src_state = src_delete(src_state);
  sf_close(outfile);
  fflush(stdout);       /* Make sure all messages are flushed */
  return 0;
}
/* ======================================================================== */

static signed int find_initial_gain(Input_file_info *info) {
/*
 * Scans the input file for the highest sample and uses the enclosing block
 * to deduce an initial gain value.  In general this doesn't work well,
 * but the code is left in as a reference point in case any further development
 * is done in future.
 *
 * Returns 0 on success or -1 on error.  The init_gain field of the
 * given Input_file_info structure will be set with the gain deduced by
 * this function.
 */
SRC_STATE *src_state = NULL;
SRC_DATA src_data;
signed int channels;
signed int i, error=0, complete=0;
float max, gain = 1.0;
int64_t frame_cx, max_sample_frame;

  printf("  Scanning for maximum amplitude... "); fflush(stdout);

  /* Take channel count from original source */
  channels = info->channels;

  /* Scan the file for the block containing the maximum peak.  Ensure we
   * start from the beginning of the file.
   */
  switch (info->filetype) {
    case INPUT_QT:
      quicktime_set_audio_position(info->input_file.qtfile,0,0);
      break;
    case INPUT_SNDFILE:
      sf_seek(info->input_file.sndfile,0,SEEK_SET);
      break;
  }
  max = -1.0;
  frame_cx = 0;
  max_sample_frame = 0;

  while (!error && !complete) {
    /* Read a block of audio data */
    switch (info->filetype) {
      case INPUT_QT: qt_read_data(info,&src_data); break;
      case INPUT_SNDFILE: 
        src_data.input_frames = sf_readf_float(info->input_file.sndfile,
          input,BUFFER_LEN/channels);
        break;
    }
    /* The last read will not be a full buffer, so flag completeness */
    if (src_data.input_frames < BUFFER_LEN/channels) {
      complete = 1;
    }
    /* Scan the audio data for the maximum sample */
    for (i=0; i<src_data.input_frames*channels; i++) {
      if (fabs(input[i]) > max) {
        max = fabs(input[i]);
        max_sample_frame = frame_cx;
      }
    }
    /* If the maximum is 1.0, stop the search since it means that the
     * current block by definition contains the maximum sample.
     */
    if (max == 1.0)
      complete = 1;
    frame_cx += src_data.input_frames;
  }
  printf("found in block at frame %lld\n", max_sample_frame);

  /* At this point max_sample_frame contains the zero-based frame number of
   * the block containing the input file's maximum sample.  Now perform the
   * sample rate conversion repeatedly on this block of data, reducing gain
   * until there is no clipping.  This will give a reasonable estimation of
   * the initial gain to use when resampling the entire file.
   */
  complete = 0;
  while (!error && !complete) {
    /* Initialise the samplerate converter.  Strange things happen if
     * this is allocated outside this loop and reused for subsequent
     * resampling attempts. */
    if ((src_state = src_new(SRC_SINC_BEST_QUALITY,channels,&error)) == NULL) {
      printf ("  Could not create sample rate converter: %s\n",src_strerror(error));
      return -1;
    }
    src_data.src_ratio = info->resample_ratio;
    src_data.data_out = output;
    src_data.output_frames = BUFFER_LEN / channels;

    max = -1.0;
    src_data.data_in = input;
    src_data.end_of_input = 1; /* We're only doing one input block */
    /* Initialise for completeness but it's not really required */
    src_data.input_frames = 0;
    /* Ensure we're at the block of interest and load the block into the 
     * input buffer.
     */
    switch (info->filetype) {
      case INPUT_QT:
        quicktime_set_audio_position(info->input_file.qtfile,max_sample_frame,0);
        qt_read_data(info,&src_data);
        break;
      case INPUT_SNDFILE:
        sf_seek(info->input_file.sndfile,max_sample_frame,SEEK_SET);
        src_data.input_frames = sf_readf_float(info->input_file.sndfile,
          input,BUFFER_LEN/channels);
        break;
    }
    src_data.data_in = input;

    /* Run the sample rate converter over the block and if all is OK, check
     * for clipping in the output.  src_process needs to be called until it
     * generates no output frames to ensure all the input data has been
     * converted.
     */
    while (!error && !complete) {
      if ((error = src_process(src_state,&src_data))) {
        printf("  Error during resampling: %s\n",src_strerror(error));
      } else
      if (src_data.output_frames_gen==0) {
        complete = 1;
      } else
      if ((max=apply_gain(src_data.data_out,src_data.output_frames_gen,channels,
           max,gain)) > 1.0) {
        /* Clipping was detected.  Reset gain, flag the clipping and break
         * out of this loop into the outer one so we can have another go.
         */
        gain = gain/max;
        complete = 0;
        break;
      }
    }
    src_state = src_delete(src_state);
  }

  /* If complete is set, the current gain is what was required to prevent
   * clipping during the resampling operation.
   */
  if (!error) {
    printf("    Best guess for initial gain: %g\n", gain);
    info->init_gain = gain;
  }
  fflush(stdout);       /* Make sure all messages are flushed */
  return 0;
}
/* ======================================================================== */

void show_usage(char *progname) {
/*
 * Tell the user how to use this program.
 */
  fprintf(stderr,
    "Audiosync version " VERSIONSTR ", (c) 2005, 2006 Jonathan Woithe\n"
    "Released under the GNU General Public License v2.\n"
    "Usage: %s [<option>]... <filename> [<filename> ...]\n"
    "Attempts to resync audio sources where each have been recorded with unlocked\n"
    "clocks.  This can occur if two unlocked digital recorders have been used, or\n"
    "where the audio and video clocks of a camcorder are free-running.\n"
    "Options:\n"
    "  -a          Try to discover a sensible value for the initial gain.  In\n"
    "              general this doesn't work well.\n"
    "  -c          Output to current directory rather than the source's directory.\n"
    "  -f=<FREQ>   Use a sampling frequency of <FREQ> Hz for the output.  Normally\n"
    "              the output sampling rate equals the input rate.\n"
    "  -g=<GAIN>   Start resampling with a gain of <GAIN> rather than 1.0.\n"
    "  -h          Display this help text.\n"
    "  -p          Only probe for length/sync information and display it.\n"
    "  -r=<R_DESC> Specify resync ratio in terms of sample numbers or the number of\n"
    "              samples in the source and destination.\n"
    "  -u          Write output as an unnormalised WAV file with 32-bit floating-\n"
    "              point samples.  Nominal range is between +/- 1.0, but peaks\n"
    "              could exceed this range due to the effects of resampling.\n\n"
    "<R_DESC> can be in one of two forms giving flexibility in specifying the\n"
    "resample ratio:\n"
    "  A-B,X-Y     resample such that (B-A+1) source samples map to (Y-X+1) samples\n"
    "              in the output stream.\n"
    "  C,Z         resample such that C samples in the source map to Z samples in\n"
    "              the output stream.\n\n"
    "The input file can be a quicktime/DV file as produced by dvgrab (for example)\n"
    "or a plain sound file.  If a sound file is given then the ratio must be\n"
    "specified using `-r'; for quicktime/DV files the ratio can be inferred from\n"
    "the lengths of the audio and video tracks.\n\n"
    "The output sampling ratio is equal to the nominal sampling rate of the input\n"
    "source (or the frequency specified with the `-f' option if given).  The\n"
    "resampling ratio thus effectively specifies the real sampling rate of the input\n"
    "relative to the nominal rate.\n",
    progname);
}
/* ======================================================================== */

double parse_resync_ratio(char *arg) {
/*
 * Parses a ratio description string from the command line.  See usage
 * for details of the formats allowed.  Return a ratio on success, or -1 on
 * error.
 */
char *str;
char *comma;
char *dash1, *dash2;
double a,b,x,y;
signed int err = 0;

  /* Need at least 3 characters for a valid specification */
  if (strlen(arg)<=2)
    return -1.0;
  str = strdup(arg);
  /* A comma is required in the expression */
  if ((comma=strchr(str,',')) == NULL)
    return -1.0;
  *comma = 0;

  dash1=strchr(str,'-');
  if (dash1 && dash1<comma) {
    /* Read two source sample numbers */
    *dash1=0;
    err = (sscanf(str,"%lg",&a)!=1 || sscanf(dash1+1,"%lg",&b)!=1);
  } else {
    /* Read a number of source samples spec */
    err = sscanf(str,"%lg",&b)!=1 || b==0;
    a=1;
  }
  dash2=strchr(comma+1,'-');
  if (dash2) {
    /* Read two destination sample numbers */
    *dash2=0;
    err = (sscanf(comma+1,"%lg",&x)!=1 || sscanf(dash2+1,"%lg",&y)!=1);
  } else {
    /* Read a number of destination samples spec */
    err = sscanf(comma+1,"%lg",&y)!=1 || y==0;
    x=1;
  }
  free(str);
  if (err || a>b || x>y)
    return -1.0;
  /* Use a,b and x,y to work out a new/old resampling frequency ratio for 
   * libsamplerate
   */
  return (1.0+y-x) / (1.0+b-a);
}
/* ======================================================================== */

int main(int argc, char *argv[]) {

signed int probe_only=0, auto_gain=0;
signed int i, have_files=0, err=0;
double resample_ratio = 0.0;
Input_file_info source_info;
char *output_name;

  memset(&source_info,0,sizeof(source_info));

  /* Scan command line arguments for options (anything beginning with "-") */
  for (i=1; i<argc && !err; i++) {
    if (argv[i][0]!='-') {
      have_files=1;
      continue;
    }
    switch (argv[i][1]) {
      case 'a':
        auto_gain=1;
        break;
      case 'c':
        cwd_output = 1;
        break;
      case 'f':
        if (argv[i][2]=='=') {
          err = (sscanf(argv[i]+3,"%d",&source_info.output_samplerate)!=1 || source_info.output_samplerate<0);
        } else
          err=1;
        if (err)
          fprintf(stderr,"Error specifying output sampling frequency: %s\n",argv[i]);
        break;
      case 'g':
        if (argv[i][2]=='=') {
          err = (sscanf(argv[i]+3,"%g",&source_info.init_gain)!=1 || source_info.init_gain<=0.0);
        } else
          err = 1;
        if (err)
          fprintf(stderr,"Error setting initial gain setting: %s\n",argv[i]);
        break;
      case 'h':
        show_usage(argv[0]);
        return 0;
        break;
      case 'p':
        probe_only=1;
        break;
      case 'r':
        if (argv[i][2]=='=') {
          err = (resample_ratio=(parse_resync_ratio(argv[i]+3))) <= 0.0;
        } else
          err = 1;
        if (err)
          fprintf(stderr,"Malformed resync ratio description: %s\n",argv[i]);
        break;
      case 'u':
        unnorm_flt = 1;
        break;
      default:
        fprintf(stderr,"Unknown option: %s\n",argv[i]);
        err=1;
    }
  }
  if (err || !have_files) {
    if (!have_files)
      fprintf(stderr,"No input files specified\n");
    return 2;
  }
  if (resample_ratio==1.0) {
    fprintf(stderr,"Resample ratio of 1.0 requested: nothing to do\n");
    return 0;
  }

  for (i = 1; i < argc; i++) {
    /* Skip command line options */
    if (argv[i][0]=='-')
      continue;
    if (!file_exists(argv[i])) {
      fprintf(stderr,"File %s does not exist\n",argv[i]);
      continue;
    }
    output_name = output_filename(argv[i]);
    source_info.resample_ratio = resample_ratio;

    if (qtfile_probe(argv[i],&source_info)==1 ||
        sndfile_probe(argv[i],&source_info)==1) {
      if (auto_gain) {
        find_initial_gain(&source_info);
      }
      if (!probe_only) {
        do_resample(&source_info,output_name);
      }
      switch (source_info.filetype) {
        case INPUT_QT: quicktime_close(source_info.input_file.qtfile); break;
        case INPUT_SNDFILE: sf_close(source_info.input_file.sndfile); break;
      }
    } else
    if (source_info.input_file.qtfile==NULL) {
      fprintf(stderr,"Format of %s unknown; nothing to do\n",argv[i]);
    }
    free(output_name);
  }
  if (input_channel_ptr)
    free(input_channel_ptr);

  return 0;
}

/* ======================================================================== */
